/* Dazuko. Allow file access control for 3rd-party applications.
   Copyright (C) 2002 H+BEDV Datentechnik GmbH
   Written by Martin Ritter <mritter@antivir.de>
              John Ogness <jogness@antivir.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/version.h>

#ifdef MODULE
#include <linux/module.h>
#endif

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#ifdef DEBUG
#define DPRINTK(x) printk x
#else
#define DPRINTK(x)
#endif

#include <linux/init.h>

#include <linux/unistd.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#ifdef __SMP__
#include <asm/smplock.h>
#endif
#include "dazuko.h"

#define NUM_SLOTS	25

#define	SCAN_ON_OPEN		(access_mask & ON_OPEN)
#define	SCAN_ON_CLOSE		(access_mask & ON_CLOSE)
#define	SCAN_ON_EXEC		(access_mask & ON_EXEC)
#define	SCAN_ON_CLOSE_MODIFIED	(access_mask & ON_CLOSE_MODIFIED)

#define	FREE	0	/* the daemon is not ready */
#define	READY	1	/* a daemon waits for something to do */
#define	WAITING	2	/* a request is waiting to be served */
#define	WORKING	3	/* daemon is currently in action */
#define	DONE	4	/* daemon response is available */

extern void 	*sys_call_table[];

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
int dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos);
#else
ssize_t dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos);
#endif
int dazuko_device_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long param);
int dazuko_device_open(struct inode *inode, struct file *file);
int dazuko_device_release(struct inode *inode, struct file *file);

struct path_t
{
	/* A node in a linked list of paths. Used
	 * for the include and exclude lists. */

	struct path_t	*next;
	int		len;
	char		path[1];	/* this MUST be at the end of the struct */
};

struct hash_t
{
	/* A node in a linked list of filenames.
	 * Used for the list of files to be
	 * scanned on close. */

	struct hash_t	*next;
	struct file	*file;
	int		dirty;
	int		namelen;
	char		name[1];	/* this MUST be at the end of the struct */
};

struct slot_t
{
	/* A representation of a daemon. It holds
	 * all information about the daemon, the
	 * file that is scanned, and the state of
	 * the scanning process. */

	int			id;		
	int			pid;		/* pid of our daemon */
	int			state;
	int			response;
	int			event;
	int			o_flags;
	int			o_mode;
	int			kuid;		/* user id of the kernel process */
	int			kpid;		/* process id of the kernel process */
	int			filenamelength;	/* not including terminator */
	char			*filename;
	struct semaphore	mutex;
};

struct dazuko_file_struct
{
	/* A structure designed for simple and
	 * intelligent memory management when
	 * doing filename lookups in the kernel. */

	const char		*user_filename;		/* userspace filename */
	int			should_scan;		/* already know we need to scan? */
	int			filename_length;	/* length of filename */
	char			*filename;		/* kernelspace filename */
	int			putname_filename;	/* flag to clean up filename */
	int			full_filename_length;	/* length of filename */
	char			*full_filename;		/* kernelspace filename with full path */
	int			free_full_filename;	/* flag to clean up full_filename */
	struct dentry		*dentry;		/* used to get inode */
	int			dput_dentry;		/* flag to clean up dentry */
	char			*buffer;		/* used to get full path */
	int			free_page_buffer;	/* flag to clean up buffer */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	struct nameidata	nd;			/* used to get full path */
	int			path_release_nd;	/* flag to clean up nd */
	struct vfsmount		*vfsmount;		/* used to get full path */
	int			mntput_vfsmount;	/* flag to clean up vfsmount */
#endif
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static struct vfsmount		*orig_rootmnt = NULL;
#endif

static struct dentry		*orig_root = NULL;
static int			active = 0;
static char			access_mask = 7;
static struct path_t		*incl_paths = NULL;
static struct path_t		*excl_paths = NULL;
static struct hash_t		*hash = NULL;
static struct slot_t		slots[NUM_SLOTS];
static int 			dev_major = -1;
static rwlock_t			lock_hash;
static rwlock_t			lock_lists;
static asmlinkage long		(*original_sys_open)(const char *filename, int flags, int mode);
static asmlinkage long		(*original_sys_close)(unsigned int fd);
static asmlinkage ssize_t	(*original_sys_write)(unsigned int fd, char *buf, unsigned int count);
static asmlinkage int		(*original_sys_execve)(struct pt_regs regs);

static struct file_operations	fops = {
					read: dazuko_device_read,	/* read */
					ioctl: dazuko_device_ioctl,	/* ioctl */
					open: dazuko_device_open,	/* open */
					release: dazuko_device_release,	/* release */
				};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static DECLARE_WAIT_QUEUE_HEAD(wait_kernel_waiting_for_free_slot);
static DECLARE_WAIT_QUEUE_HEAD(wait_kernel_waiting_while_daemon_works);
static DECLARE_WAIT_QUEUE_HEAD(wait_daemon_waiting_for_work);
static DECLARE_WAIT_QUEUE_HEAD(wait_daemon_waiting_for_free);
#else
static struct wait_queue *wait_kernel_waiting_for_free_slot;
static struct wait_queue *wait_kernel_waiting_while_daemon_works;
static struct wait_queue *wait_daemon_waiting_for_work;
static struct wait_queue *wait_daemon_waiting_for_free;


/* The following code is taken directly from Linux in the file:
   include/linux/sched.h */

#ifndef __wait_event_interruptible
#define __wait_event_interruptible(wq, condition, ret)		 \
do {								 \
	struct wait_queue __wait;				 \
								 \
	__wait.task = current;					 \
	add_wait_queue(&wq, &__wait);				 \
	for (;;) {						 \
		current->state = TASK_INTERRUPTIBLE;		 \
		mb();						 \
		if (condition)					 \
			break;					 \
		if (!signal_pending(current)) {			 \
			schedule();				 \
			continue;				 \
		}						 \
		ret = -ERESTARTSYS;				 \
		break;						 \
	}							 \
	current->state = TASK_RUNNING;				 \
	remove_wait_queue(&wq, &__wait);			 \
} while (0)
#endif

#ifndef wait_event_interruptible
#define wait_event_interruptible(wq, condition)			 \
({								 \
	int __ret = 0;						 \
	if (!(condition))					 \
		__wait_event_interruptible(wq, condition, __ret);\
	__ret;							 \
})
#endif

#endif

void dazuko_bzero(void *p, int len)
{
	/* "zero out" len bytes starting with p */

	char	*ptr = (char *)p;

	while (len--)
		*ptr++ = 0;
}

int slot_state(int slot_num)
{
	int state;

/* DOWN */
	if (down_interruptible(&slots[slot_num].mutex) != 0)
		return -EINTR;

	state = slots[slot_num].state;

	up(&slots[slot_num].mutex);
/* UP */

	return state;
}

int __dazuko_change_slot_state(int slot_num, int from_state, int to_state)
{
	/* Make a predicted state transition. We fail if it
	 * is an unpredicted change. We can ALWAYS go to the
	 * to_state if it is the same as from_state. Not SMP safe! */

	if (to_state != from_state)
	{
		/* make sure this is a predicted transition and there
		 * is a daemon on this slot (pid != 0)*/
		if (slots[slot_num].state != from_state || !slots[slot_num].pid)
			return 0;
	}

	slots[slot_num].state = to_state;

	/* handle appropriate wake_up's for basic
	 * state changes */

	if (to_state == READY)
	{
		wake_up(&wait_kernel_waiting_for_free_slot);
	}
	else if (to_state == FREE)
	{
		wake_up(&wait_kernel_waiting_while_daemon_works);
		wake_up(&wait_daemon_waiting_for_free);
	}

	return 1;
}

int dazuko_change_slot_state(int slot_num, int from_state, int to_state, int release)
{
	/* SMP safe version of __dazuko_change_slot_state().
	 * This should only be used if we haven't
	 * already aquired slot.mutex. Use this function
	 * with CAUTION, since the mutex may or may not
	 * be released depending on the return value AND
	 * on the value of the "release" argument. */

	int	success;

	/* if we are interrupted, report the state as unpredicted */
/* DOWN */
	if (down_interruptible(&slots[slot_num].mutex) != 0)
		return 0;

	success = __dazuko_change_slot_state(slot_num, from_state, to_state);

	/* the mutex is released if the state change was
	 * unpredicted or if the called wants it released */
	if (!success || release)
		up(&slots[slot_num].mutex);
/* UP */
	return success;
}

int dazuko_find_slot(int pid, int release)
{
	/* Find the first slot with the same given
	 * pid number. SMP safe. Use this function
	 * with CAUTION, since the mutex may or may not
	 * be released depending on the return value AND
	 * on the value of the "release" argument. */

	int	i;

	for (i=0 ; i<NUM_SLOTS ; i++)
	{
/* DOWN */
		/* if we are interrupted, we say that no
		 * slot was found */
		if (down_interruptible(&slots[i].mutex) != 0)
			return -1;

		if (slots[i].pid == pid)
		{
			/* we release the mutex only if the
			 * called wanted us to */
			if (release)
				up(&slots[i].mutex);
/* UP */
			return i;
		}

		up(&slots[i].mutex);
/* UP */
	}

	return -1;
}

int dazuko_insert_path_fs(struct path_t **list, char *fs_path, int fs_len)
{
	/* Create a new path_t structure and insert it
	 * into the linked list given (list argument).
	 * The fs_len argument is to help speed things
	 * up so we don't have to calculate the length
	 * of fs_path. The fs_path argument is in
	 * userspace! */

	struct path_t	*newitem;
	struct path_t	*tmp;

	/* create a new path_t structure making room for path also */
	newitem = (struct path_t *)kmalloc(sizeof(struct path_t) + fs_len, GFP_KERNEL);
	if (!newitem)
		return -EFAULT;

	/* We must copy the path from userspace to kernelspace. */

	if (copy_from_user(newitem->path, fs_path, fs_len) != 0)
	{
		kfree(newitem);
		return -EFAULT;
	}

	newitem->path[fs_len] = 0;

	while (newitem->path[fs_len] == 0)
	{
		fs_len--;
		if (fs_len == 0)
			break;
	}

	newitem->len = fs_len;

	/* we want only absolute paths */
	if (newitem->path[0] != '/')
	{
		kfree(newitem);
		return -EINVAL;
	}

	/* check if this path already exists in the list */
	for (tmp=*list ; tmp ; tmp=tmp->next)
	{
		if (newitem->len == tmp->len)
		{
			if (memcmp(newitem->path, tmp->path, tmp->len) == 0)
			{
				/* we already have this path */

				kfree(newitem);

				return 0;
			}
		}
	}

	DPRINTK(("dazuko: adding %s %s\n", (list == &incl_paths) ? "incl" : "excl", newitem->path));

	/* add path_t to head of linked list */
/* LOCK */
	write_lock(&lock_lists);
	newitem->next = *list;
	*list = newitem;
	write_unlock(&lock_lists);
/* UNLOCK */

	return 0;
}

int dazuko_device_open(struct inode *inode, struct file *file)
{
	/* We register the daemon by finding an
	 * unused slot (slot->pid=0) and setting
	 * the slot->pid to the proccess id of
	 * the current proccess, the daemon. */

	int	i;

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return 0;

	/* find a slot with pid 0 and hold the mutex
	 * if we find one */
/* DOWN? */
	i = dazuko_find_slot(0, 0);

	if (i < 0)
		return -EBUSY;

/* DOWN */

	/* We have found a slot, so increment the active
	 * variable and the kernel module use counter.
	 * The module counter will always reflect the
	 * number of daemons. */

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
	active++;

	slots[i].pid = current->pid;

	/* the daemon is registered, but not yet
	 * ready to receive files */
	__dazuko_change_slot_state(i, FREE, FREE);

	DPRINTK(("dazuko: slot[%d] assigned to daemon %d\n", i, current->pid));

	up(&slots[i].mutex);
/* UP */
	return 0;
}

void dazuko_remove_all_hash(void)
{
	/* Empty the hash linked list. */

	struct hash_t	*tmp;

/* LOCK */
	write_lock(&lock_hash);
	while (hash)
	{
		tmp = hash;
		hash = hash->next;

		kfree(tmp);
	}
	write_unlock(&lock_hash);
/* UNLOCK */
}

void dazuko_remove_all_paths(void)
{
	/* Empty both include and exclude path_t
	 * linked lists. */

	struct path_t	*tmp;

/* LOCK */
	write_lock(&lock_lists);

	/* empty include paths list */
	while (incl_paths)
	{
		tmp = incl_paths;
		incl_paths = incl_paths->next;

		DPRINTK(("dazuko: removing incl %s\n", tmp->path));

		kfree(tmp);
	}

	/* empty exclude paths list */
	while (excl_paths)
	{
		tmp = excl_paths;
		excl_paths = excl_paths->next;

		DPRINTK(("dazuko: removing excl %s\n", tmp->path));

		kfree(tmp);
	}

	write_unlock(&lock_lists);
/* UNLOCK */
}

int dazuko_device_release(struct inode *inode, struct file *file)
{
	/* We unregister the daemon by finding the
	 * slot with the same slot->pid as the the
	 * current process id, the daemon. */

	int		i;

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return 0;

	/* find our slot and hold the mutex
	 * if we find it */
/* DOWN? */
	i = dazuko_find_slot(current->pid, 0);

	if (i < 0)
	{
		DPRINTK(("dazuko: daemon %d had no slot (possible bug)\n", current->pid));
		return 0;
	}

/* DOWN */

	/* active should always be positive here, but
	 * let's check just to be sure. ;) */
	if (active > 0)
	{
		/* active and the kernel usage counter
		 * should always reflect how many daemons
		 * are active */

#ifdef MODULE
		MOD_DEC_USE_COUNT;
#endif
		active--;
	}
	else
	{
		DPRINTK(("dazuko: active count error (possible bug)\n"));
	}

	/* clearing the pid makes the slot available */
	slots[i].pid = 0;

	/* reset slot state */
	__dazuko_change_slot_state(i, FREE, FREE);

	up(&slots[i].mutex);
/* UP */

	/* Wake up any kernel processes that are
	 * waiting for an available slot. Remove
	 * all the include and exclude paths
	 * if there are no more deamons */
	if (active == 0)
	{
		/* clear out include and exclude paths */
		dazuko_remove_all_paths();

		/* clear out hash nodes */
		dazuko_remove_all_hash();

		wake_up(&wait_kernel_waiting_for_free_slot);
	}

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
int dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos)
#else
ssize_t dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos)
#endif
{
	/* Reading from the dazuko device simply
	 * returns the device number. This is to
	 * help out the daemon. */

	char	tmp[10];
	int	dev_major_len;

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return 0;

	if (dev_major < 0)
		return -ENODEV;

	/* print dev_major to a string
	 * and get length (with terminator) */
	dev_major_len = sprintf(tmp, "%d", dev_major) + 1;

	if (length < dev_major_len)
		return -EINVAL;

	/* copy dev_major string to userspace */
	if (copy_to_user(buffer, tmp, dev_major_len) != 0)
		return -EFAULT;

	return dev_major_len;
}

int dazuko_device_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long param)
{
	/* A daemon uses this function to interact with
	 * the kernel. A daemon can set scanning parameters,
	 * give scanning response, and get filenames to scan. */

	int		error;
	int		i;
	int		len;
	struct access_t	*u_dazuko;

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return 0;

	/* A macro is used to translate the cmd argument.
	 * This keeps it compatible across various Linux
	 * platforms. */
	switch (_IOC_NR(cmd))
	{
		case IOCTL_GET_AN_ACCESS:
			/* The daemon is requesting a filename of a file
			 * to scan. This code will wait until a filename
			 * is available, or until we should be killed.
			 * (killing is done if any errors occur as well
			 * as when the user kills us) */

			error = verify_area(VERIFY_WRITE, (void *)param, sizeof(struct access_t));
			if (error)
			{
				return error;
			}

			u_dazuko = (struct access_t *)param;

tryagain:
			/* find our slot */
			i = dazuko_find_slot(current->pid, 1);

			if (i < 0)
			{
				return -ESRCH;
			}

			/* the daemon is now ready to receive a file */
			dazuko_change_slot_state(i, READY, READY, 1);

			if (wait_event_interruptible(wait_daemon_waiting_for_work, slot_state(i) != READY) != 0)
			{
				/* The user has issued an interrupt.
				 * Return an error. The daemon should
				 * unregister itself. */

				DPRINTK(("dazuko: daemon %d killed while waiting for work\n", current->pid));

				return -EINTR;
			}

			/* slot SHOULD now be in WAITING state */

			/* we will be writing data to the slot, so
			 * we need to lock it */
/* DOWN */
			if (down_interruptible(&slots[i].mutex) != 0)
			{
				return -EINTR;
			}

			if (!__dazuko_change_slot_state(i, WAITING, WORKING))
			{
				/* State transition error. Try again., */

				up(&slots[i].mutex);
/* UP */
				goto tryagain;
			}

			/* Slot IS in WORKING state. Copy all the
			 * necessary information to userspace structure. */

			if (copy_to_user(u_dazuko->filename, slots[i].filename, slots[i].filenamelength+1) != 0)
			{
				up(&slots[i].mutex);
/* UP */
				return -EFAULT;
			}

			if (copy_to_user(&u_dazuko->event, &slots[i].event, sizeof(int)) != 0)
			{
				up(&slots[i].mutex);
/* UP */
				return -EFAULT;
			}

			if (copy_to_user(&u_dazuko->o_flags, &slots[i].o_flags, sizeof(int)) != 0)
			{
				up(&slots[i].mutex);
/* UP */
				return -EFAULT;
			}

			if (copy_to_user(&u_dazuko->o_mode, &slots[i].o_mode, sizeof(int)) != 0)
			{
				up(&slots[i].mutex);
/* UP */
				return -EFAULT;
			}

			if (copy_to_user(&u_dazuko->uid, &slots[i].kuid, sizeof(int)) != 0)
			{
				up(&slots[i].mutex);
/* UP */
				return -EFAULT;
			}

			if (copy_to_user(&u_dazuko->pid, &slots[i].kpid, sizeof(int)) != 0)
			{
				up(&slots[i].mutex);
/* UP */
				return -EFAULT;
			}

			up(&slots[i].mutex);
/* UP */
			return 0;  /* no error */

		case IOCTL_RETURN_ACCESS:
			/* The daemon has finished scanning a file
			 * and has the response to give. The daemon's
			 * slot should be in the WORKING state. */

			error = verify_area(VERIFY_READ, (void *)param, sizeof(struct access_t));
			if (error)
			{
				return error;
			}

			u_dazuko = (struct access_t *)param;

			/* find our slot */
			i = dazuko_find_slot(current->pid, 1);

			if (i < 0)
			{
				/* It appears the kernel isn't interested
				* in us or our response. It gave our slot away! */

				DPRINTK(("dazuko: daemon %d unexpectedly lost slot[%d]\n", current->pid, i));

				return -EPERM;
			}

			/* we will be writing into the slot, so we
			 * need to lock it */
/* DOWN */
			if (down_interruptible(&slots[i].mutex) != 0)
			{
				return -EINTR;
			}

			if (!__dazuko_change_slot_state(i, WORKING, DONE))
			{
				/* The slot is in the wrong state. We will
				 * assume the kernel has cancelled the file
				 * access. */

				DPRINTK(("dazuko: response from daemon %d on slot[%d] not needed\n", current->pid, i));

				up(&slots[i].mutex);
/* UP */
				return 0;
			}

			/* copy the response into the slot */
			if (copy_from_user(&slots[i].response, &(u_dazuko->deny), sizeof(int)) != 0)
			{
				up(&slots[i].mutex);
/* UP */
				return -EFAULT;
			}

			up(&slots[i].mutex);
/* UP */

			/* wake up any kernel processes that are
			 * waiting for responses */
			wake_up(&wait_kernel_waiting_while_daemon_works);

			if (wait_event_interruptible(wait_daemon_waiting_for_free, slot_state(i) != DONE) != 0)
			{
				/* The user has issued an interrupt.
				 * Return an error. The daemon should
				 * unregister itself. */

				DPRINTK(("dazuko: daemon %d killed while waiting for response acknowledgement\n", current->pid));

				return -EINTR;
			}

			return 0;

		case IOCTL_SET_OPTION:
			/* The daemon wants to set a configuration
			 * option in the kernel. */

			error = verify_area(VERIFY_READ, (void *)param, 2*sizeof(int));
			if (error)
				return error;

			/* copy option type from userspace */
			if (copy_from_user(&i, (int *)param, sizeof(int)) != 0)
				return -EPERM;

			param += sizeof(int);

			/* copy path length from userspace */
			if (copy_from_user(&len, (int *)param, sizeof(int)) != 0)
				return -EPERM;

			param += sizeof(int);

			error = verify_area(VERIFY_READ, (void *)param, len);
			if (error)
				return error;

			/* check option type and take the appropriate action */
			switch (i)
			{
				case SET_ACCESS_MASK:
					if (copy_from_user(&access_mask, (char *)param, sizeof(char)) != 0)
						return -EFAULT;
					break;

				case ADD_INCLUDE_PATH:
					dazuko_insert_path_fs(&incl_paths, (char *)param, len);
					break;

				case ADD_EXCLUDE_PATH:
					dazuko_insert_path_fs(&excl_paths, (char *)param, len);
					break;

				case REMOVE_ALL_PATHS:
					dazuko_remove_all_paths();
					break;

				default:
					DPRINTK(("dazuko: daemon %d requested unknown set %d (possible bug)\n", current->pid, i));
					break;
			}
			break;
		default:
			DPRINTK(("dazuko: daemon %d requested unknown device_ioctl %d (possible bug)\n", current->pid, _IOC_NR(cmd)));
			break;
	}

	return 0;
}

int dazuko_get_and_hold_ready_slot(void)
{
	/* This is a simple search to find a
	 * slot whose state is READY. This means
	 * it is able to accept work. If a slot
	 * is found, the slot.mutex is held so
	 * it can be filled with work by the caller.
	 * It is the responsibility of the caller
	 * to RELEASE THE MUTEX. */

	int	i;

	for (i=0 ; i<NUM_SLOTS ; i++)
	{
/* DOWN? */
		if (dazuko_change_slot_state(i, READY, WAITING, 0))
		{
/* DOWN */
			return i;
		}
	}

	/* we didn't find a slot that is ready for work */

	return -1;
}

int dazuko_run_daemon(int event, char *filename, int filenamelength, int o_flags, int o_mode)
{
	/* This is the main function called by the kernel
	 * to work with a daemon. */

	int	rc;
	int	i;
	int	pid;

begin:
	/* we initialize the slot value because
	 * we cannot guarentee that it will be
	 * assigned a new value BEFORE !active
	 * is checked */
	i = -1;

	/* wait for a slot to become ready */
	if (wait_event_interruptible(wait_kernel_waiting_for_free_slot, ((i = dazuko_get_and_hold_ready_slot()) >= 0) || !active) != 0)
	{
		/* The kernel process was killed while
		 * waiting for a slot to become ready.
		 * This is fine. */

		return -1;  /* user interrupted */
	}

	/* Make sure we have a slot. We may have
	 * gotten past the last wait because we
	 * are no longer active. */

	if (i < 0)
	{
		/* We were no longer active. We don't
		 * need to initiate a daemon. This also
		 * means we never acquired the lock. */

		return 0;  /* allow access */
	}

/* DOWN */

	/* the slot is already locked at this point */

	/* grab the daemon's pid */
	pid = slots[i].pid;

	/* At this point we have a locked slot. It IS
	 * sitting in the WAITING state, waiting for
	 * us to give it some work. */
	
	/* set up the slot to do work */
	slots[i].filename = filename;
	slots[i].event = event;
	slots[i].response = 0;
	slots[i].kuid = current->uid;
	slots[i].kpid = current->pid;
	slots[i].o_flags = o_flags;
	slots[i].o_mode = o_mode;
	slots[i].filenamelength = filenamelength;

	/* we are done modifying the slot */
	up(&slots[i].mutex);
/* UP */

	/* wake up any daemons waiting for work */
	wake_up(&wait_daemon_waiting_for_work);

	/* wait until the daemon is finished with the slot */
	if (wait_event_interruptible(wait_kernel_waiting_while_daemon_works, slot_state(i) != WAITING && slot_state(i) != WORKING) != 0)
	{
		/* The kernel process was killed while
		 * waiting for a daemon to process the file.
		 * This is fine. */

		/* change the slot's state to let the
		 * daemon know we are not interested
		 * in a response */
		dazuko_change_slot_state(i, FREE, FREE, 1);

		return -1;  /* user interrupted */
	}

	/* we are working with the slot, so
	 * we need to lock it */
/* DOWN */
	if (down_interruptible(&slots[i].mutex) != 0)
	{
		return -1;  /* user interrupted */
	}

	/* make sure this is the right daemon */
	if (slots[i].pid != pid)
	{
		/* This is a different daemon than
		 * the one we assigned work to.
		 * We need to scan again. */
		up(&slots[i].mutex);
/* UP */
		goto begin;
	}

	/* The slot should now be in the DONE state. */
	if (!__dazuko_change_slot_state(i, DONE, FREE))
	{
		/* The daemon was killed while scanning.
		 * We need to scan again. */

		up(&slots[i].mutex);
/* UP */
		goto begin;
	}

	/* grab the response */
	rc = slots[i].response;

	up(&slots[i].mutex);
/* UP */

	/* CONGRATULATIONS! You successfully completed a full state cycle! */

	return rc;
}

int dazuko_is_our_daemon(void)
{
	/* Check if the current process is one
	 * of the daemons. */

	return (dazuko_find_slot(current->pid, 1) >= 0);
}

int dazuko_is_selected(char *filename, int len)
{
	/* Check if the given filename (with path) is
	 * under our include directories but not under
	 * the exclude directories. */

	struct path_t	*path;

	/* If we are interrupted here, we will report that
	 * this file is not selected. This will make the
	 * kernel allow normal access. Is this dangerous? */
/* LOCK */
	read_lock(&lock_lists);

	/* check if filename is under our include paths */
	for (path=incl_paths ; path ; path=path->next)
	{
		/* the include item must be at least as long as the given filename */
		if (path->len < len)
		{
			/* the include item should match the beginning of the given filename */
			if (memcmp(path->path, filename, path->len) == 0)
				break;
		}
	}

	/* If we didn't find a path, it isn't in our
	 * include directories. It can't be one of
	 * the selected files to scan. */
	if (!path)
	{
		read_unlock(&lock_lists);
/* UNLOCK */
		return 0;
	}

	/* check if filename is under our exclude paths */
	for (path=excl_paths ; path ; path=path->next)
	{
		/* the exclude item must be at least as long as the given filename */
		if (path->len < len)
		{
			/* the exclude item should match the beginning of the given filename */
			if (memcmp(path->path,filename,path->len) == 0)
				break;
		}
	}

	read_unlock(&lock_lists);
/* UNLOCK */

	/* If we got a path, then we are supposed
	 * to exclude this file for scanning. */
	if (path)
		return 0;

	/* if we made it this far, it is a selected file to scan */

	return 1;
}

int dazuko_add_hash(struct file *file, char *filename, int len)
{
	/* Add the given file and filename to the linked list
	 * of files to scan once they are closed. */

	struct hash_t	*h;

	/* create a new hash_t structure making room for name also */
	h = (struct hash_t *)kmalloc(sizeof(struct hash_t) + len, GFP_KERNEL);
	if (!h)	
		return -EFAULT;

	/* fill in structure items */

	h->file = file;
	h->dirty = 0;
	h->namelen = len;
	memcpy(h->name, filename, len);
	h->name[len] = 0;

	/* add the new hash_t item to the head of the
	 * hast_t linked list */

/* LOCK */
	write_lock(&lock_hash);
	h->next = hash;
	hash = h;
	write_unlock(&lock_hash);
/* UNLOCK */
	return 0;
}

struct hash_t *dazuko_find_hash(struct file *file)
{
	/* Find the given file within our list
	 * of files for scanning after closing. */

	struct hash_t	*h;

/* LOCK */
	read_lock(&lock_hash);

	for (h=hash ; h ; h=h->next)
	{
		/* since these are file* and not
		 * strings, we can compare them
		 * directly */
		if (h->file == file)
			break;
	} 

	read_unlock(&lock_hash);
/* UP */

	return h;
}

struct hash_t *dazuko_get_hash(struct file *file)
{
	/* Find the given file within our list
	 * and then remove it from the list and
	 * return it. */

	struct hash_t	*prev;
	struct hash_t	*cur;

/* LOCK */
	write_lock(&lock_hash);

	prev = NULL;
	cur = hash;
	while (cur)
	{
		/* since these are file* and not
		 * strings, we can compare them
		 * directly */
		if (cur->file == file)
		{
			/* remove the item from the list */
			if (!prev)
				hash = cur->next;
			else
				prev->next = cur->next;
			break;
		}

		prev = cur;
		cur = cur->next;
	}

	write_unlock(&lock_hash);
/* UNLOCK */

	return cur;
}

int dazuko_get_filename_length(char *filename)
{
	/* Get the length of the filename. There is
	 * currently a DAZUKO_FILENAME_MAX_LENGTH maximum size restriction
	 * on filenames. :( */

	int len;

	for (len=0 ; len<DAZUKO_FILENAME_MAX_LENGTH && filename[len]; len++);

	if (len == DAZUKO_FILENAME_MAX_LENGTH)
	{
		DPRINTK(("dazuko: filename too long (%s)\n", filename));

		filename[DAZUKO_FILENAME_MAX_LENGTH] = 0;
	}

	return len;
}

int dazuko_get_dentry(struct dazuko_file_struct *kfs)
{
	/* We get the appropriate structures in order
	 * to acquire the inode and store them in the
	 * dazuko_file_struct structure. */

	/* make sure we really need to get the filename */
	if (!kfs->putname_filename)
	{
		/* grab filename from filename cache */
		kfs->filename = (char *)getname(kfs->user_filename);

		/* make sure it is a valid name */
		if (IS_ERR(kfs->filename))
			return 0;

		/* the name will need to be put back */
		kfs->putname_filename = 1;
	}

	/* get filename length and make sure it isn't too long */
	kfs->filename_length = dazuko_get_filename_length(kfs->filename);
	if (kfs->filename_length == DAZUKO_FILENAME_MAX_LENGTH)
		return 0;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		dazuko_bzero(&(kfs->nd), sizeof(struct nameidata));

		/* initialize nameidata structure for finding file data */
		if (!path_init(kfs->filename, LOOKUP_FOLLOW | LOOKUP_POSITIVE, &(kfs->nd)))
			return 0;

		if (!kfs->path_release_nd)
		{
			/* find file data and fill it in nameidata structure */
			if (path_walk(kfs->filename, &(kfs->nd)))  /* !=0 -> error */
				return 0;

			/* the nameidata will need to be released */
			kfs->path_release_nd = 1;
		}

		/* get a local copy of the dentry to make kernel version
		 * compatibility code eaiser to read */

		/* make sure we don't already have a dentry */
		if (!kfs->dput_dentry)
		{
			kfs->dentry = dget(kfs->nd.dentry);

			/* the dentry will need to be put back */
			kfs->dput_dentry = 1;
		}
	}
	#else
	{
		if (!kfs->dput_dentry)
		{
			kfs->dentry = lookup_dentry(kfs->filename, NULL, 1);
			if (IS_ERR(kfs->dentry))
				return 0;

			/* the dentry will need to be put back */
			kfs->dput_dentry = 1;
		}
	}
	#endif

	/* check if this file has no inode */
	if (kfs->dentry->d_inode == NULL)
		return 0;

	/* if we made it this far, we got the inode */

	return 1;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
char * __d_path(struct dentry *dentry, struct dentry *root, char *buffer, int buflen)
{
	/* Copy of d_path from linux/dcache.c but using
	 * a given root instead of the current root. */

	char * end = buffer+buflen;
	char * retval;

	*--end = '\0';
	buflen--;
	if (dentry->d_parent != dentry && list_empty(&dentry->d_hash)) {
		buflen -= 10;
		end -= 10;
		memcpy(end, " (deleted)", 10);
	}

	/* Get '/' right */
	retval = end-1;
	*retval = '/';

	for (;;) {
		struct dentry * parent;
		int namelen;

		if (dentry == root)
			break;
		dentry = dentry->d_covers;
		parent = dentry->d_parent;
		if (dentry == parent)
			break;
		namelen = dentry->d_name.len;
		buflen -= namelen + 1;
		if (buflen < 0)
			break;
		end -= namelen;
		memcpy(end, dentry->d_name.name, namelen);
		*--end = '/';
		retval = end;
		dentry = parent;
	}
	return retval;
}
#endif

int dazuko_get_full_filename(struct dazuko_file_struct *kfs)
{
	/* Get the filename with the full path appended
	 * to the beginning. */

	char		*temp;
	struct dentry	*root;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		struct vfsmount	*rootmnt;
	#endif

	/* check if we need to allocate a buffer */
	if (!kfs->free_page_buffer)
	{
		/* get pre-requisites for d_path function */
		kfs->buffer = (char *)__get_free_page(GFP_USER);

		/* the buffer will need to be freed */
		kfs->free_page_buffer = 1;
	}

	root = dget(orig_root);

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		/* make sure we don't already have a vfsmount */
		if (!kfs->mntput_vfsmount)
		{
			kfs->vfsmount = mntget(kfs->nd.mnt);

			/* the vfsmount will need to be put back */
			kfs->mntput_vfsmount = 1;
		}

		/* build new filename with path included, using temp */

		rootmnt = mntget(orig_rootmnt);

		spin_lock(&dcache_lock);
		temp = __d_path(kfs->dentry, kfs->vfsmount, root, rootmnt, kfs->buffer, PAGE_SIZE);
		spin_unlock(&dcache_lock);

		mntput(rootmnt);
	}
	#else
	{
		/* build new filename with path included, using temp */

		temp = __d_path(kfs->dentry, root, kfs->buffer, PAGE_SIZE);
	}
	#endif

	dput(root);

	/* make sure we really got a new filename */
	if (!temp)
		return 0;

	/* make sure we don't already have a full_filename */
	if (!kfs->free_full_filename)
	{
		/* get new filename length and make sure it isn't too long */
		kfs->full_filename_length = dazuko_get_filename_length(temp);
		if (kfs->full_filename_length == DAZUKO_FILENAME_MAX_LENGTH)
			return 0;

		kfs->full_filename = (char *)kmalloc(kfs->full_filename_length + 1, GFP_KERNEL);

		/* the char array will need to be freed */
		kfs->free_full_filename = 1;

		memcpy(kfs->full_filename, temp, kfs->full_filename_length + 1);
	}

	/* we have a filename with the full path */

	return 1;
}

void dazuko_file_struct_critical_cleanup(struct dazuko_file_struct *kfs)
{
	/* Delete all the flagged structures from the
	 * given dazuko_file_struct and reset all critical
	 * values back to 0. */

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		if (kfs->mntput_vfsmount)
		{
			mntput(kfs->vfsmount);
			kfs->mntput_vfsmount = 0;
		}
	}
	#endif

	if (kfs->free_page_buffer)
	{
		free_page((unsigned long)kfs->buffer);
		kfs->free_page_buffer = 0;
	}

	if (kfs->dput_dentry)
	{
		dput(kfs->dentry);
		kfs->dput_dentry = 0;
	}

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		if (kfs->path_release_nd)
		{
			path_release(&(kfs->nd));
			kfs->path_release_nd = 0;
		}
	}
	#endif

	if (kfs->putname_filename)
	{
		putname(kfs->filename);
		kfs->putname_filename = 0;
	}
}

void dazuko_file_struct_cleanup(struct dazuko_file_struct *kfs)
{
	kfs->should_scan = 0;

	dazuko_file_struct_critical_cleanup(kfs);

	if (kfs->free_full_filename)
	{
		kfree(kfs->full_filename);
		kfs->free_full_filename = 0;
	}
}

int dazuko_should_scan(struct dazuko_file_struct *kfs)
{
	/* Check if we are supposed to scan this file.
	 * This checks for all the correct file types,
	 * permissions, and if it is within the desired
	 * paths to scan. */

	int success = 0;

	/* check if we already know we scan this file */
	if (kfs->should_scan)
	{
		return 1;
	}

	/* make sure we can get an inode */
	if (dazuko_get_dentry(kfs))
	{
		/* make sure we have a regular file */
		if (S_ISREG(kfs->dentry->d_inode->i_mode))
		{
			/* make sure the file is readable */
			if (permission(kfs->dentry->d_inode, MAY_READ) == 0)
			{
				/* make sure we can get the full path */
				if (dazuko_get_full_filename(kfs))
				{
					/* check if the filename is within our include
	 				* directories but not our exclude directories */

					if (dazuko_is_selected(kfs->full_filename, kfs->full_filename_length))
					{
						/* If we made it this far, we are supposed
				 		* to scan this file. We mark it so that
				 		* any further immediate inquiries don't have
				 		* to do all this work all over again. */

						kfs->should_scan = 1;

						success = 1;
					}
				}
			}
		}
	}

	dazuko_file_struct_critical_cleanup(kfs);

	return success;
}

asmlinkage int dazuko_sys_execve(struct pt_regs regs)
{
	/* The kernel wants to execute the given file.
	 * Because the given structure contains stack
	 * address information, we can't simply call
	 * the default standard execve. Instead we
	 * have to manually inline the standard execve
	 * call. */

	struct dazuko_file_struct	kfs;
	char			*filename;
	int			error = 0;

	/* check if we are supposed to do scanning */
	if (!active || !SCAN_ON_EXEC)
		goto standard;

	/* do not scan if it is our scan daemon
	 * exeing the file */
	if (dazuko_is_our_daemon())
	{
		goto standard;
	}

	/* start with a clean dazuko_file_struct */
	dazuko_bzero(&kfs, sizeof(struct dazuko_file_struct));

	kfs.user_filename = (char *)regs.ebx;

	/* make sure we should scan this file */
	if (dazuko_should_scan(&kfs))
		error = dazuko_run_daemon(ON_EXEC, kfs.full_filename, kfs.full_filename_length, 0, 0);

	if (error > 0)
	{
		/* virus found and not cleaned */

		return -EPERM;
	}
	else if (error < 0)
	{
		/* user interrupted */

		return -EINTR;
	}

	dazuko_file_struct_cleanup(&kfs);

	/* call the standard execve function */

	/* We cannot simply call the original version of execvc
	 * because the parameter contains stack information and
	 * the call with push the execvc call onto a new stack
	 * level and seg fault. :( */

standard:
	/* The following code only works on i386 machines.
	 * It is directly copied from Linux in the file:
	 * arch/i386/kernel/process.c */

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		filename = getname((char *) regs.ebx);
		error = PTR_ERR(filename);
		if (IS_ERR(filename))
			goto out;
		error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, &regs);
		if (error == 0)
			current->ptrace &= ~PT_DTRACE;
		putname(filename);
out:
		return error;
	}
	#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,20)
	{
		#ifdef __SMP__
			lock_kernel();
		#endif
		filename = getname((char *) regs.ebx);
		error = PTR_ERR(filename);
		if (IS_ERR(filename))
			goto out;
		error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, &regs);
		if (error == 0)
			current->ptrace &= ~PT_DTRACE;
		putname(filename);
out:
		#ifdef __SMP__
			unlock_kernel();
		#endif
		return error;
	}
	#else
	{
		#ifdef __SMP__
			lock_kernel();
		#endif
		filename = getname((char *) regs.ebx);
		error = PTR_ERR(filename);
		if (IS_ERR(filename))
			goto out;
		error = do_execve(filename, (char **) regs.ecx, (char **) regs.edx, &regs);
		if (error == 0)
			current->flags &= ~PF_DTRACE;
		putname(filename);
out:
		#ifdef __SMP__
			unlock_kernel();
		#endif
		return error;
	}
	#endif


}

asmlinkage long	dazuko_sys_open(const char *filename, int flags, int mode)
{
	/* The kernel wants to open the given filename
	 * with the given flags and mode. The dazuko_file_struct
	 * is used to handle the tricky job of cleaning
	 * up the many pieces of memory that may or may
	 * not be allocated. */

	struct dazuko_file_struct	kfs;
	int			error = 0;
	int			fd;

	/* Check if we are supposed to do scanning. Even
	 * if we don't scan on open, we need to keep going
	 * if we are supposed to scan on close. */
	if (!active || filename == NULL || !(SCAN_ON_OPEN | SCAN_ON_CLOSE | SCAN_ON_CLOSE_MODIFIED))
	{
		return original_sys_open(filename, flags, mode);
	}

	/* do not scan if it is our scan daemon
	 * opening the file */
	if (dazuko_is_our_daemon())
	{
		return original_sys_open(filename, flags, mode);
	}

	/* start with a clean dazuko_file_struct */
	dazuko_bzero(&kfs, sizeof(struct dazuko_file_struct));

	kfs.user_filename = filename;

	/* make sure we are supposed to scan files on open and
	* that we aren't truncating this file on open (truncating
	* the file will delete any contents, so no worry for viruses. */
	if (SCAN_ON_OPEN)
	{
		/* make sure we should scan this file */
		if (dazuko_should_scan(&kfs))
		{
			error = dazuko_run_daemon(ON_OPEN, kfs.full_filename, kfs.full_filename_length, flags, mode);
		}
	}

	if (error > 0)
	{
		/* virus found and not cleaned */

		fd = -EPERM;
	}
	else if (error < 0)
	{
		/* user interrupted */

		fd = -EINTR;
	}
	else
	{
		/* call the standard open function */
		fd = original_sys_open(filename, flags, mode);

		/* if the file was opened and we are interested
	 	* in scanning on close, add this file to our hash_t list */

		if (active && fd > 0 && fd < NR_OPEN)
		{
			if (SCAN_ON_CLOSE || (SCAN_ON_CLOSE_MODIFIED && (flags & (O_RDWR | O_WRONLY))))
			{
				/* make sure we should scan this file */
				if (dazuko_should_scan(&kfs))
				{
					#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
					{
						read_lock(&current->files->file_lock);
					}
					#endif

					dazuko_add_hash(current->files->fd[fd], kfs.full_filename, kfs.full_filename_length);

					#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
					{
						read_unlock(&current->files->file_lock);
					}
					#endif
				}
			}
		}
	}

	dazuko_file_struct_cleanup(&kfs);

	return fd;
}

asmlinkage long	dazuko_sys_close(unsigned int fd)
{
	/* The kernel wants to close the given file
	 * descriptor. */

	int		error;
	struct hash_t	*h = NULL;
	struct file	*file = NULL;

	/* do not scan if it is our scan daemon
	 * closing the file */
	if (dazuko_is_our_daemon())
	{
		return original_sys_close(fd);
	}

	/* If it is a valid file descriptor, see if it is
	 * in our list of files to scan on close. If it is,
	 * it will be removed from the list also. */

	if (active && (SCAN_ON_CLOSE || SCAN_ON_CLOSE_MODIFIED) && fd > 0 && fd < NR_OPEN)
	{
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		{
			read_lock(&current->files->file_lock);
		}
		#endif

		/* grab the file* for possible later use */
		file = current->files->fd[fd];

		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		{
			read_unlock(&current->files->file_lock);
		}
		#endif
	}

	/* call the standard close function */
	error = original_sys_close(fd);

	if (!error && active && (SCAN_ON_CLOSE || SCAN_ON_CLOSE_MODIFIED) && fd > 0 && fd < NR_OPEN)
	{
		/* find hash entry and remove it from list */
		h = dazuko_get_hash(file);

		/* if we found the file in our list and the file was
	 	* successfully closed, we need to scan it */
		if (h)
		{
			/* determine if we are scanning on close and/or close_modified */

			/* note that modified has priority over just close */

			if (SCAN_ON_CLOSE_MODIFIED && h->dirty)
				dazuko_run_daemon(ON_CLOSE_MODIFIED, h->name, h->namelen, 0, 0);
			else if (SCAN_ON_CLOSE)
				dazuko_run_daemon(ON_CLOSE, h->name, h->namelen, 0, 0);

			/* clean up the hash_t structure */
			kfree(h);
		}
	}

	return error;
}

asmlinkage ssize_t dazuko_sys_write(unsigned int fd, char *buf, unsigned int count)
{
	/* The kernel wants to write to the given file
	 * descriptor. */

	int		num;
	struct hash_t	*h = NULL;

	/* do not track if it is our scan daemon
	 * writing the file */
	if (dazuko_is_our_daemon())
	{
		return original_sys_write(fd, buf, count);
	}

	/* Check if this file is in our list of files to
	 * be cleaned on close. It will not be removed.
	 * We only have to do this if we are scanning on
	 * close.*/

	if (active && SCAN_ON_CLOSE_MODIFIED && fd > 0 && fd < NR_OPEN)
	{
		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		{
			read_lock(&current->files->file_lock);
		}
		#endif

		h = dazuko_find_hash(current->files->fd[fd]);

		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		{
			read_unlock(&current->files->file_lock);
		}
		#endif
	}
	
	/* call the standard write function */
	num = original_sys_write(fd, buf, count);
	
	/* if we actually wrote something and we found the
	 * file in our list, set it as dirty */

	if (num > 0 && h)
		h->dirty = 1;

	return num;
}


int __init dazuko_init(void)
{
	/* Called insmod when inserting the module. */

	int	i;

	/* "zero out" all the slots */
	dazuko_bzero(&slots, sizeof(slots));

	/* give each slot a unique id */
	for (i=0 ; i<NUM_SLOTS ; i++)
	{
		slots[i].id = i;
		#ifdef init_MUTEX
			init_MUTEX(&slots[i].mutex);
		#else
			sema_init(&slots[i].mutex, 1);
		#endif
	}

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		rwlock_init(&lock_hash);
		rwlock_init(&lock_lists);
	}
	#else
	{
		lock_hash = RW_LOCK_UNLOCKED;
		lock_lists = RW_LOCK_UNLOCKED;
	}
	#endif

	/* Make sure we have a valid task_struct. */

	if (current == NULL)
	{
		printk("dazuko: panic (current == NULL)\n");
		return -1;
	}
	if (current->fs == NULL)
	{
		printk("dazuko: panic (current->fs == NULL)\n");
		return -1;
	}
	if (current->fs->root == NULL)
	{
		printk("dazuko: panic (current->root == NULL)\n");
		return -1;
	}
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		if (current->fs->rootmnt == NULL)
		{
			printk("dazuko: panic (current->rootmnt == NULL)\n");
			return -1;
		}
	}
	#endif

	/* register the dazuko device */
	dev_major = register_chrdev(0, DEVICE_NAME, &fops);
	if (dev_major < 0)
	{
		printk("dazuko: unable to register device chrdev, err=%d\n", dev_major);
		return dev_major;
	}

	/* Grab the current root. This is assumed to be the real.
	 * If it is not the real root, we could have problems
	 * looking up filenames. */

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	{
		read_lock(&current->fs->lock);
		orig_rootmnt = current->fs->rootmnt;
	}
	#endif

	orig_root = current->fs->root;

	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		read_unlock(&current->fs->lock);
	#endif

	/* do a file syncronization on all devices (IMPORTANT!) and replace system calls */
	#ifdef __SMP__
		lock_kernel();
	#endif

	fsync_dev(0);

	/* replace the system call entries with our entries */
	original_sys_open = sys_call_table[__NR_open];
	sys_call_table[__NR_open] = dazuko_sys_open;
	
	original_sys_close = sys_call_table[__NR_close];
	sys_call_table[__NR_close] = dazuko_sys_close;

	original_sys_write = sys_call_table[__NR_write];
	sys_call_table[__NR_write] = dazuko_sys_write;

	original_sys_execve = sys_call_table[__NR_execve];
	sys_call_table[__NR_execve] = dazuko_sys_execve;

	#ifdef __SMP__
		unlock_kernel();
	#endif
	/* done syncing and replacing */

	/* initialization complete */

	printk("Dazuko: loaded, version=%s, dev_major=%d\n", VERSION, dev_major);

  	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
void __exit dazuko_exit(void)
#else
void dazuko_exit(void)
#endif
{
	/* Called by rmmod when removing the module. */

	int		error;

	dazuko_remove_all_paths();
	dazuko_remove_all_hash();

	/* do a file syncronization on all devices (IMPORTANT!) and replace system calls */
	#ifdef __SMP__
		lock_kernel();
	#endif

	fsync_dev(0);

	/* check if it is still our entries in the sytem call table */
  	if (sys_call_table[__NR_open] != dazuko_sys_open
			|| sys_call_table[__NR_close] != dazuko_sys_close
			|| sys_call_table[__NR_write] != dazuko_sys_write
			|| sys_call_table[__NR_execve] != dazuko_sys_execve)
    		printk("dazuko: open/close/write/execve system calls not correct (system may be left in an unstable state!)\n");

	/* return original system calls (we HOPE no one has played with the table) */

	sys_call_table[__NR_open] = original_sys_open;
	sys_call_table[__NR_close] = original_sys_close;
	sys_call_table[__NR_write] = original_sys_write;
	sys_call_table[__NR_execve] = original_sys_execve;

	#ifdef __SMP__
		unlock_kernel();
	#endif
	/* done syncing and replacing */

	error = unregister_chrdev(dev_major, DEVICE_NAME);
	if (error < 0)
	{
		printk("dazuko: error unregistering chrdev, err=%d\n", error);
	}

	DPRINTK(("dazuko: module unloaded\n"));
}

#ifdef MODULE

int init_module(void)
{
	return dazuko_init();
}

void cleanup_module(void)
{
	dazuko_exit();
}

MODULE_AUTHOR("H+BEDV Datentechnik GmbH <linux_support@antivir.de>");
MODULE_DESCRIPTION("allow 3rd-party file access control");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,10)
MODULE_LICENSE("GPL");
#endif

EXPORT_NO_SYMBOLS;

#else

module_init(dazuko_init);
module_exit(dazuko_exit);
//module_init(int dazuko_init(void));
//module_exit(void dazuko_exit(void));

#endif
