/* Example program demonstrating the capabilities/interface of Dazuko.
   Copyright (C) 2002 H+BEDV Datentechnik GmbH
   Written by John Ogness <jogness@antivir.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* This program is a simple application demonstrating how to interface
   with Dazuko. The program instructs Dazuko to detect open, close, and
   execute accesses within the /home directory (or any subdirectory
   thereof). The program than prints out the accesses and instructs
   Dazuko to allow access.

   An application communicates with Dazuko through a device. After
   installing Dazuko (as kernel module or after building in the
   kernel itself), the device node must be created. The major device
   number is needed. This can be found by looking at /proc/devices.

   example:
   > grep dazuko /proc/devices
   < 254 dazuko
   > mknod -m 666 /dev/dazuko c 254 0

   After Dazuko is loaded and the device node is created, this example
   program SHOULD run (as is) without any problems.

   Note: For security reasons, Dazuko will only interact with applications
         running as root. Therefore, this example program must also be run
         as root.
 */

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include "dazukoio.h"


int	RUNNING = 1;	/* flag for main loop */

void sigterm(int sig)
{
	/* exit the loop on TERM signal */

	printf("detected TERM signal\n");

	RUNNING = 0;
	signal(sig, sigterm);
}

void print_access(struct access_t *acc_ptr)
{
	/* print access data */

	switch (acc_ptr->event)
	{
		case ON_OPEN:
			printf("OPEN ");
			break;
		case ON_CLOSE:
			printf("CLOSE ");
			break;
		case ON_CLOSE_MODIFIED:
			printf("CLOSE (modified) ");
			break;
		case ON_EXEC:
			printf("EXEC ");
			break;
		default:
			printf("???? ");
			break;
	}
	printf("flags: %d ", acc_ptr->o_flags);
	printf("mode: %d ", acc_ptr->o_mode);
	printf("uid: %d ", acc_ptr->uid);
	printf("pid: %d ", acc_ptr->pid);
	printf("file: %s ", acc_ptr->filename);
	printf("\n");
}

int main(void)
{
	struct access_t	acc;

	/* make sure we are root */
	if (getuid() != 0)
	{
		printf("only root can run this program\n");
		return -1;
	}

	/* register with dazuko */
	if (dazukoRegister() != 0)
	{
		printf("error: failed to register with Dazuko\n");
		return -1;
	}

	printf("registerd with Dazuko successfully\n");

	/* detect TERM signals */
	signal(15, sigterm);

	/* set access mask */
	if (dazukoSetAccessMask(ON_OPEN | ON_CLOSE | ON_CLOSE_MODIFIED | ON_EXEC) != 0)
	{
		printf("error: failed to set access mask\n");
		dazukoUnregister();
		return -1;
	}

	printf("set access mask successfully\n");

	/* set scan path */
	if (dazukoAddIncludePath("/home") != 0)
	{
		printf("error: failed to add include path\n");
		dazukoUnregister();
		return -1;
	}

	printf("set scan path successfully\n");

	while (RUNNING)
	{
		/* get an access */
		if (dazukoGetAccess(&acc) == 0)
		{
			print_access(&acc);

			/* always allow access */
			acc.deny = 0;

			/* return access (IMPORTANT, the kernel is waiting for us!) */
			if (dazukoReturnAccess(&acc) != 0)
			{
				printf("error: failed to return access\n");
				RUNNING = 0;
			}
		}
		else
		{
			printf("warning: failed to get an access\n");
		}
	}

	/* unregister with Dazuko */
	dazukoUnregister();

	return 0;
}
